@@ -55,10 +55,8 @@ static int send_file_control_msg(int fd, int type, void *buf, int len)
     if (buf)
         memcpy(msg.buf, buf, len);
 
-    if (write(fd, &msg, sizeof(msg)) < 0) {
-        log_err("write fifo: %s\n", strerror(errno));
+    if (write(fd, &msg, sizeof(msg)) < 0)
         return -1;
-    }
 
     return 0;
 }
@@ -76,14 +74,21 @@ void file_context_reset(struct file_context *ctx)
     }
 
     ctx->busy = false;
+
+    if (ctx->buf) {
+        free(ctx->buf);
+        ctx->buf = NULL;
+    }
 }
 
-static void notify_user_canceled(struct rtty *rtty)
+static void notify_user_canceled(struct tty *tty)
 {
+    struct rtty *rtty = tty->rtty;
+
     buffer_put_u8(&rtty->wb, MSG_TYPE_FILE);
-    buffer_put_u16be(&rtty->wb, 2);
-    buffer_put_u8(&rtty->wb, rtty->file_context.sid);
-    buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_CANCELED);
+    buffer_put_u16be(&rtty->wb, 33);
+    buffer_put_data(&rtty->wb, tty->file.sid, 32);
+    buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_ABORT);
     ev_io_start(rtty->loop, &rtty->iow);
 }
 
@@ -97,14 +102,23 @@ static int notify_progress(struct file_context *ctx)
 
 static void send_file_data(struct file_context *ctx)
 {
-    struct rtty *rtty = container_of(ctx, struct rtty, file_context);
-    uint8_t buf[4096 * 4];
+    struct tty *tty = container_of(ctx, struct tty, file);
+    struct rtty *rtty = tty->rtty;
     int ret;
 
+    if (!ctx->buf) {
+        ctx->buf = malloc(UPLOAD_FILE_BUF_SIZE);
+        if (!ctx->buf) {
+            log_err("malloc: %s\n", strerror(errno));
+            send_file_control_msg(ctx->ctlfd, RTTY_FILE_MSG_ERR, NULL, 0);
+            goto err;
+        }
+    }
+
     if (ctx->fd < 0)
         return;
 
-    ret = read(ctx->fd, buf, sizeof(buf));
+    ret = read(ctx->fd, ctx->buf, UPLOAD_FILE_BUF_SIZE);
     if (ret < 0) {
         send_file_control_msg(ctx->ctlfd, RTTY_FILE_MSG_ERR, NULL, 0);
         goto err;
@@ -113,10 +127,10 @@ static void send_file_data(struct file_context *ctx)
     ctx->remain_size -= ret;
 
     buffer_put_u8(&rtty->wb, MSG_TYPE_FILE);
-    buffer_put_u16be(&rtty->wb, 2 + ret);
-    buffer_put_u8(&rtty->wb, ctx->sid);
+    buffer_put_u16be(&rtty->wb, 33 + ret);
+    buffer_put_data(&rtty->wb, ctx->sid, 32);
     buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_DATA);
-    buffer_put_data(&rtty->wb, buf, ret);
+    buffer_put_data(&rtty->wb, ctx->buf, ret);
     ev_io_start(rtty->loop, &rtty->iow);
 
     if (ret == 0) {
@@ -130,13 +144,14 @@ static void send_file_data(struct file_context *ctx)
     return;
 
 err:
-    notify_user_canceled(rtty);
+    notify_user_canceled(tty);
     file_context_reset(ctx);
 }
 
 static int start_upload_file(struct file_context *ctx, const char *path)
 {
-    struct rtty *rtty = container_of(ctx, struct rtty, file_context);
+    struct tty *tty = container_of(ctx, struct tty, file);
+    struct rtty *rtty = tty->rtty;
     const char *name = basename(path);
     struct stat st;
     int fd;
@@ -150,9 +165,9 @@ static int start_upload_file(struct file_context *ctx, const char *path)
     fstat(fd, &st);
 
     buffer_put_u8(&rtty->wb, MSG_TYPE_FILE);
-    buffer_put_u16be(&rtty->wb, 2 + strlen(name));
-    buffer_put_u8(&rtty->wb, ctx->sid);
-    buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_INFO);
+    buffer_put_u16be(&rtty->wb, 33 + strlen(name));
+    buffer_put_data(&rtty->wb, ctx->sid, 32);
+    buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_SEND);
     buffer_put_string(&rtty->wb, name);
     ev_io_start(rtty->loop, &rtty->iow);
 
@@ -165,9 +180,10 @@ static int start_upload_file(struct file_context *ctx, const char *path)
     return 0;
 }
 
-bool detect_file_operation(uint8_t *buf, int len, int sid, struct file_context *ctx)
+bool detect_file_operation(uint8_t *buf, int len, const char *sid, struct file_context *ctx)
 {
-    struct rtty *rtty = container_of(ctx, struct rtty, file_context);
+    struct tty *tty = container_of(ctx, struct tty, file);
+    struct rtty *rtty = tty->rtty;
     char fifo_name[128];
     pid_t pid;
     int ctlfd;
@@ -208,14 +224,14 @@ bool detect_file_operation(uint8_t *buf, int len, int sid, struct file_context *
         return true;
     }
 
-    ctx->sid = sid;
     ctx->ctlfd = ctlfd;
+    strcpy(ctx->sid, sid);
 
     if (buf[3] == 'R') {
         buffer_put_u8(&rtty->wb, MSG_TYPE_FILE);
-        buffer_put_u16be(&rtty->wb, 2);
-        buffer_put_u8(&rtty->wb, ctx->sid);
-        buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_START_DOWNLOAD);
+        buffer_put_u16be(&rtty->wb, 33);
+        buffer_put_data(&rtty->wb, ctx->sid, 32);
+        buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_RECV);
         ev_io_start(rtty->loop, &rtty->iow);
 
         send_file_control_msg(ctlfd, RTTY_FILE_MSG_REQUEST_ACCEPT, NULL, 0);
@@ -261,7 +277,6 @@ bool detect_file_operation(uint8_t *buf, int len, int sid, struct file_context *
 
 static void start_download_file(struct file_context *ctx, struct buffer *info, int len)
 {
-    struct rtty *rtty = container_of(ctx, struct rtty, file_context);
     char *name = savepath + strlen(savepath);
     struct mntent *ment;
     struct statvfs sfs;
@@ -285,6 +300,12 @@ static void start_download_file(struct file_context *ctx, struct buffer *info, i
 
     buffer_pull(info, name, len - 4);
 
+    if (!access(savepath, F_OK)) {
+        send_file_control_msg(ctx->ctlfd, RTTY_FILE_MSG_ERR_EXIST, NULL, 0);
+        log_err("the file '%s' already exists\n", name);
+        goto open_fail;
+    }
+
     fd = open(savepath, O_WRONLY | O_TRUNC | O_CREAT, 0644);
     if (fd < 0) {
         send_file_control_msg(ctx->ctlfd, RTTY_FILE_MSG_ERR, NULL, 0);
@@ -312,22 +333,23 @@ static void start_download_file(struct file_context *ctx, struct buffer *info, i
 check_space_fail:
     buffer_pull(info, name, len - 4);
 open_fail:
-    notify_user_canceled(rtty);
     file_context_reset(ctx);
 }
 
-static void send_file_data_ack(struct rtty *rtty)
+static void send_file_data_ack(struct tty *tty)
 {
+    struct rtty *rtty = tty->rtty;
+
     buffer_put_u8(&rtty->wb, MSG_TYPE_FILE);
-    buffer_put_u16be(&rtty->wb, 2);
-    buffer_put_u8(&rtty->wb, rtty->file_context.sid);
-    buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_DATA_ACK);
+    buffer_put_u16be(&rtty->wb, 33);
+    buffer_put_data(&rtty->wb, tty->file.sid, 32);
+    buffer_put_u8(&rtty->wb, RTTY_FILE_MSG_ACK);
     ev_io_start(rtty->loop, &rtty->iow);
 }
 
 void parse_file_msg(struct file_context *ctx, struct buffer *data, int len)
 {
-    struct rtty *rtty = container_of(ctx, struct rtty, file_context);
+    struct tty *tty = container_of(ctx, struct tty, file);
     int type = buffer_pull_u8(data);
 
     len--;
@@ -344,13 +366,12 @@ void parse_file_msg(struct file_context *ctx, struct buffer *data, int len)
                 ctx->remain_size -= len;
 
                 if (notify_progress(ctx) < 0) {
-                    notify_user_canceled(rtty);
                     file_context_reset(ctx);
                 } else {
                     if (ctx->remain_size == 0)
                         file_context_reset(ctx);
                     else
-                        send_file_data_ack(rtty);
+                        send_file_data_ack(tty);
                 }
             } else {
                 buffer_pull(data, NULL, len);
@@ -360,20 +381,13 @@ void parse_file_msg(struct file_context *ctx, struct buffer *data, int len)
         }
         break;
 
-    case RTTY_FILE_MSG_DATA_ACK:
+    case RTTY_FILE_MSG_ACK:
         send_file_data(ctx);
         break;
 
-    case RTTY_FILE_MSG_CANCELED:
-        if (ctx->fd > -1) {
-            close(ctx->fd);
-            ctx->fd = -1;
-        }
-
-        send_file_control_msg(ctx->ctlfd, RTTY_FILE_MSG_CANCELED, NULL, 0);
-        close(ctx->ctlfd);
-
-        ctx->busy = false;
+    case RTTY_FILE_MSG_ABORT:
+        send_file_control_msg(ctx->ctlfd, RTTY_FILE_MSG_ABORT, NULL, 0);
+        file_context_reset(ctx);
         break;
 
     default:
