@@ -29,7 +29,7 @@
 #include <sys/sysinfo.h>
 
 #include "net.h"
-#include "web.h"
+#include "http.h"
 #include "file.h"
 #include "rtty.h"
 #include "list.h"
@@ -46,34 +46,34 @@ static void del_tty(struct tty *tty)
 
     ev_io_stop(loop, &tty->ior);
     ev_io_stop(loop, &tty->iow);
+    ev_timer_stop(loop, &tty->tmr);
     ev_child_stop(loop, &tty->cw);
 
+    rtty->ntty--;
+    list_del(&tty->node);
+
     buffer_free(&tty->wb);
 
     close(tty->pty);
     kill(tty->pid, SIGTERM);
 
-    rtty->ttys[tty->sid] = NULL;
+    file_context_reset(&tty->file);
 
-    file_context_reset(&rtty->file_context);
-
-    log_info("delete tty: %d\n", tty->sid);
+    log_info("delete tty: %s\n", tty->sid);
 
     free(tty);
 }
 
-static inline struct tty *find_tty(struct rtty *rtty, int sid)
+static inline struct tty *find_tty(struct rtty *rtty, const char *sid)
 {
-    if (sid > RTTY_MAX_TTY - 1)
-        return NULL;
-    return rtty->ttys[sid];
-}
+    struct tty *tty;
 
-static inline void tty_logout(struct rtty *rtty, int sid)
-{
-    struct tty *tty = find_tty(rtty, sid);
-    if (tty)
-        del_tty(tty);
+    list_for_each_entry(tty, &rtty->ttys, node) {
+        if (!strcmp(tty->sid, sid))
+            return tty;
+    }
+
+    return NULL;
 }
 
 static void pty_on_read(struct ev_loop *loop, struct ev_io *w, int revents)
@@ -82,7 +82,9 @@ static void pty_on_read(struct ev_loop *loop, struct ev_io *w, int revents)
     struct rtty *rtty = tty->rtty;
     struct buffer *wb = &rtty->wb;
     static uint8_t buf[4096];
-    int len;
+    int len = 0;
+
+    tty->active = ev_now(loop);
 
     while (1) {
         len = read(w->fd, buf, sizeof(buf));
@@ -101,12 +103,18 @@ static void pty_on_read(struct ev_loop *loop, struct ev_io *w, int revents)
             return;
     }
 
-    if (detect_file_operation(buf, len, tty->sid, &rtty->file_context))
+    if (detect_file_operation(buf, len, tty->sid, &tty->file))
         return;
 
+    tty->wait_ack += len;
+
+    /* stop until received ack */
+    if (tty->wait_ack > RTTY_TTY_ACK_BLOCK)
+        ev_io_stop(loop, w);
+
     buffer_put_u8(wb, MSG_TYPE_TERMDATA);
-    buffer_put_u16be(wb, len + 1);
-    buffer_put_u8(wb, tty->sid);
+    buffer_put_u16be(wb, 32 + len);
+    buffer_put_data(wb, tty->sid, 32);
     buffer_put_data(wb, buf, len);
     ev_io_start(loop, &rtty->iow);
 }
@@ -134,45 +142,73 @@ static void pty_on_exit(struct ev_loop *loop, struct ev_child *w, int revents)
     struct buffer *wb = &rtty->wb;
 
     buffer_put_u8(wb, MSG_TYPE_LOGOUT);
-    buffer_put_u16be(wb, 1);
-    buffer_put_u8(wb, tty->sid);
+    buffer_put_u16be(wb, 32);
+    buffer_put_data(wb, tty->sid, 32);
     ev_io_start(loop, &rtty->iow);
 
     del_tty(tty);
 }
 
-static void tty_login(struct rtty *rtty)
+static void tty_timer_cb(struct ev_loop *loop, struct ev_timer *w, int revents)
 {
-    struct tty *tty;
+    struct tty *tty = container_of(w, struct tty, tmr);
+    ev_tstamp now = ev_now(loop);
+
+    if (now - tty->active < RTTY_TTY_TIMEOUT)
+        return;
+
+    ev_timer_stop(loop, w);
+    kill(tty->pid, SIGTERM);
+
+    log_err("tty(%s) inactive over %ds, now kill it\n", tty->sid, RTTY_TTY_TIMEOUT);
+}
+
+static void tty_login(struct rtty *rtty, const char *sid)
+{
+    struct tty *tty = NULL;
+    int code = 1;
     pid_t pid;
     int pty;
-    int sid;
 
     buffer_put_u8(&rtty->wb, MSG_TYPE_LOGIN);
+    buffer_put_u16be(&rtty->wb, 33);
+    buffer_put_data(&rtty->wb, sid, 32);
 
-    for (sid = 0; sid < RTTY_MAX_TTY; sid++) {
-        if (!rtty->ttys[sid])
-            break;
-    }
-
-    if (sid == RTTY_MAX_TTY) {
+    if (rtty->ntty == RTTY_MAX_TTY) {
         log_info("tty login fail, device busy\n");
-        buffer_put_u16be(&rtty->wb, 1);
-        buffer_put_u8(&rtty->wb, 1);
-        ev_io_start(rtty->loop, &rtty->iow);
-        return;
+        goto done;
     }
 
     tty = calloc(1, sizeof(struct tty));
+    if (!tty) {
+        log_err("calloc: %s\n", strerror(errno));
+        goto done;
+    }
 
     pid = forkpty(&pty, NULL, NULL, NULL);
-    if (pid == 0)
-        rtty->username ? execl(login_path, "-p", "-f", rtty->username, NULL) : execl(login_path, login_path, NULL);
+    if (pid < 0) {
+        log_err("forkpty: %s\n", strerror(errno));
+        goto done;
+    }
+
+    if (pid == 0) {
+        if (rtty->username)
+            execl(login_path, "login", "-f", rtty->username, NULL);
+        else
+            execl(login_path, "login", NULL);
+
+        exit(1);
+    }
 
-    tty->sid = sid;
     tty->pid = pid;
     tty->pty = pty;
     tty->rtty = rtty;
+    tty->file.fd = -1;
+
+    strcpy(tty->sid, sid);
+
+    rtty->ntty++;
+    list_add(&tty->node, &rtty->ttys);
 
     fcntl(pty, F_SETFL, fcntl(pty, F_GETFL, 0) | O_NONBLOCK);
 
@@ -184,39 +220,37 @@ static void tty_login(struct rtty *rtty)
     ev_child_init(&tty->cw, pty_on_exit, pid, 0);
     ev_child_start(rtty->loop, &tty->cw);
 
-    rtty->ttys[sid] = tty;
+    ev_timer_init(&tty->tmr, tty_timer_cb, 3, 3);
+    ev_timer_start(rtty->loop, &tty->tmr);
 
-    buffer_put_u16be(&rtty->wb, 2);
-    buffer_put_u8(&rtty->wb, 0);
-    buffer_put_u8(&rtty->wb, sid);
+    code = 0;
+
+    log_info("new tty: %d/%d %s\n", rtty->ntty, RTTY_MAX_TTY, sid);
+
+done:
+    if (code)
+        free(tty);
+
+    buffer_put_u8(&rtty->wb, code);
     ev_io_start(rtty->loop, &rtty->iow);
-
-    log_info("new tty: %d\n", sid);
 }
 
-static void write_data_to_tty(struct rtty *rtty, int sid, int len)
+static void write_data_to_tty(struct tty *tty, int len)
 {
-    struct tty *tty = find_tty(rtty, sid);
-    if (!tty) {
-        log_err("non-existent sid: %d\n", sid);
-        return;
-    }
+    struct rtty *rtty = tty->rtty;
+
+    tty->active = ev_now(rtty->loop);
 
     buffer_put_data(&tty->wb, buffer_data(&rtty->rb), len);
     buffer_pull(&rtty->rb, NULL, len);
     ev_io_start(rtty->loop, &tty->iow);
 }
 
-static void set_tty_winsize(struct rtty *rtty, int sid)
+static void set_tty_winsize(struct tty *tty)
 {
-    struct tty *tty = find_tty(rtty, sid);
+    struct rtty *rtty = tty->rtty;
     struct winsize size = {};
 
-    if (!tty) {
-        log_err("non-existent sid: %d\n", sid);
-        return;
-    }
-
     size.ws_col = buffer_pull_u16be(&rtty->rb);
     size.ws_row = buffer_pull_u16be(&rtty->rb);
 
@@ -226,6 +260,8 @@ static void set_tty_winsize(struct rtty *rtty, int sid)
 
 void rtty_exit(struct rtty *rtty)
 {
+    struct tty *tty, *ntty;
+
     if (rtty->sock < 0)
         return;
 
@@ -235,13 +271,15 @@ void rtty_exit(struct rtty *rtty)
     buffer_free(&rtty->rb);
     buffer_free(&rtty->wb);
 
-    for (int i = 0; i < RTTY_MAX_TTY; i++)
-        if (rtty->ttys[i])
-            del_tty(rtty->ttys[i]);
+    list_for_each_entry_safe(tty, ntty, &rtty->ttys, node) {
+        del_tty(tty);
+    }
 
 #ifdef SSL_SUPPORT
-    if (rtty->ssl)
+    if (rtty->ssl) {
         ssl_session_free(rtty->ssl);
+        rtty->ssl_negotiated = false;
+    }
 #endif
 
     if (rtty->sock > 0) {
@@ -249,7 +287,7 @@ void rtty_exit(struct rtty *rtty)
         rtty->sock = -1;
     }
 
-    web_reqs_free(&rtty->web_reqs);
+    http_conns_free(&rtty->http_conns);
 
     if (!rtty->reconnect)
         ev_break(rtty->loop, EVBREAK_ALL);
@@ -257,7 +295,7 @@ void rtty_exit(struct rtty *rtty)
 
 static void rtty_register(struct rtty *rtty)
 {
-    size_t len = 3 + strlen(rtty->devid);
+    size_t len = 4 + strlen(rtty->devid);
     struct buffer *wb = &rtty->wb;
 
     if (rtty->description)
@@ -269,6 +307,8 @@ static void rtty_register(struct rtty *rtty)
     buffer_put_u8(wb, MSG_TYPE_REGISTER);
     buffer_put_u16be(wb, len);
 
+    buffer_put_u8(wb, RTTY_PROTO_VER);
+
     buffer_put_string(wb, rtty->devid);
     buffer_put_u8(wb, '\0');
 
@@ -283,6 +323,50 @@ static void rtty_register(struct rtty *rtty)
     ev_io_start(rtty->loop, &rtty->iow);
 }
 
+static void parse_tty_msg(struct rtty *rtty, int type, int len)
+{
+    struct buffer *b = &rtty->rb;
+    struct tty *tty = NULL;
+    char sid[33] = "";
+
+    buffer_pull(b, sid, 32);
+    len -= 32;
+
+    if (type != MSG_TYPE_LOGIN) {
+        tty = find_tty(rtty, sid);
+        if (!tty) {
+            log_err("non-existent sid: %s\n", sid);
+            buffer_pull(&rtty->rb, NULL, len);
+            return;
+        }
+    }
+
+    switch (type) {
+    case MSG_TYPE_LOGIN:
+        tty_login(rtty, sid);
+        break;
+    case MSG_TYPE_LOGOUT:
+        del_tty(tty);
+        break;
+    case MSG_TYPE_TERMDATA:
+        write_data_to_tty(tty, len);
+        break;
+    case MSG_TYPE_WINSIZE:
+        set_tty_winsize(tty);
+        break;
+    case MSG_TYPE_FILE:
+        parse_file_msg(&tty->file, b, len);
+        break;
+    case MSG_TYPE_ACK:
+        tty->wait_ack -= buffer_pull_u16be(b);
+        ev_io_start(rtty->loop, &tty->ior);
+        break;
+    default:
+        /* never to here */
+        break;
+    }
+}
+
 static int parse_msg(struct rtty *rtty)
 {
     struct buffer *rb = &rtty->rb;
@@ -298,6 +382,11 @@ static int parse_msg(struct rtty *rtty)
             return 0;
 
         msgtype = buffer_pull_u8(rb);
+        if (msgtype > MSG_TYPE_MAX) {
+            log_err("invalid message type: %d\n", msgtype);
+            return -1;
+        }
+
         buffer_pull_u16(rb);
 
         switch (msgtype) {
@@ -313,19 +402,12 @@ static int parse_msg(struct rtty *rtty)
             break;
 
         case MSG_TYPE_LOGIN:
-            tty_login(rtty);
-            break;
-
         case MSG_TYPE_LOGOUT:
-            tty_logout(rtty, buffer_pull_u8(rb));
-            break;
-
         case MSG_TYPE_TERMDATA:
-            write_data_to_tty(rtty, buffer_pull_u8(rb), msglen - 1);
-            break;
-
         case MSG_TYPE_WINSIZE:
-            set_tty_winsize(rtty, buffer_pull_u8(rb));
+        case MSG_TYPE_FILE:
+        case MSG_TYPE_ACK:
+            parse_tty_msg(rtty, msgtype, msglen);
             break;
 
         case MSG_TYPE_CMD:
@@ -336,17 +418,12 @@ static int parse_msg(struct rtty *rtty)
         case MSG_TYPE_HEARTBEAT:
             break;
 
-        case MSG_TYPE_FILE:
-            parse_file_msg(&rtty->file_context, rb, msglen);
-            break;
-
-        case MSG_TYPE_WEB:
-            web_request(rtty, msglen);
+        case MSG_TYPE_HTTP:
+            http_request(rtty, msglen);
             break;
 
         default:
-            log_err("invalid message type: %d\n", msgtype);
-            buffer_pull(rb, NULL, msglen);
+            /* never to here */
             break;
         }
     }
@@ -355,24 +432,32 @@ static int parse_msg(struct rtty *rtty)
 #ifdef SSL_SUPPORT
 static void on_ssl_verify_error(int error, const char *str, void *arg)
 {
+    bool *valid_cert = arg;
+
+    *valid_cert = false;
+
     log_warn("SSL certificate error(%d): %s\n", error, str);
 }
 
 /* -1 error, 0 pending, 1 ok */
 static int ssl_negotiated(struct rtty *rtty)
 {
+    bool valid_cert = true;
     char err_buf[128];
     int ret;
 
-    ret = ssl_connect(rtty->ssl, false, on_ssl_verify_error, NULL);
+    ret = ssl_connect(rtty->ssl, false, on_ssl_verify_error, &valid_cert);
     if (ret == SSL_PENDING)
         return 0;
 
     if (ret == SSL_ERROR) {
-        log_err("ssl connect error(%d): %s\n", ssl_err_code, ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
+        log_err("ssl connect error: %s\n", ssl_last_error_string(err_buf, sizeof(err_buf)));
         return -1;
     }
 
+    if (!valid_cert && !rtty->insecure)
+        return -1;
+
     rtty->ssl_negotiated = true;
 
     return 1;
@@ -386,8 +471,7 @@ static int rtty_ssl_read(int fd, void *buf, size_t count, void *arg)
 
     ret = ssl_read(rtty->ssl, buf, count);
     if (ret == SSL_ERROR) {
-        log_err("ssl_read(%d): %s\n", ssl_err_code,
-                ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
+        log_err("ssl_read: %s\n", ssl_last_error_string(err_buf, sizeof(err_buf)));
         return P_FD_ERR;
     }
 
@@ -433,7 +517,7 @@ static void on_net_read(struct ev_loop *loop, struct ev_io *w, int revents)
         goto err;
 
     if (eof) {
-        log_info("socket closed by server\n");
+        log_err("socket closed by server\n");
         goto err;
     }
 
@@ -463,8 +547,7 @@ static void on_net_write(struct ev_loop *loop, struct ev_io *w, int revents)
 
         ret = ssl_write(rtty->ssl, buffer_data(b), buffer_length(b));
         if (ret == SSL_ERROR) {
-            log_err("ssl_write(%d): %s\n", ssl_err_code,
-                    ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
+            log_err("ssl_write: %s\n", ssl_last_error_string(err_buf, sizeof(err_buf)));
             goto err;
         }
 
@@ -516,6 +599,7 @@ static void on_net_connected(int sock, void *arg)
         if (!rtty->ssl) {
             log_err("SSL session create fail\n");
             ev_break(rtty->loop, EVBREAK_ALL);
+            return;
         }
         ssl_set_server_name(rtty->ssl, rtty->host);
 #endif
@@ -587,9 +671,8 @@ int rtty_start(struct rtty *rtty)
             && !rtty->reconnect)
         return -1;
 
-    rtty->file_context.fd = -1;
-
-    INIT_LIST_HEAD(&rtty->web_reqs);
+    INIT_LIST_HEAD(&rtty->ttys);
+    INIT_LIST_HEAD(&rtty->http_conns);
 
     rtty->active = ev_now(rtty->loop);
 
